from string import Template

from ..data_managers.semantic_header import ConstructedNodes, NodeConstructor, Node, \
    RelationConstructor, RecordConstructor
from ..database_managers.db_connection import Query


class SemanticHeaderQueryLibrary:
    @staticmethod
    def get_create_node_by_record_constructor_query(node_constructor: NodeConstructor, batch_size: int = 5000, merge=True) -> Query:
        # find events that contain the entity as property and not nan
        # save the value of the entity property as id and also whether it is a virtual entity
        # create a new entity node if it not exists yet with properties
        merge_or_create = 'MERGE' if merge else 'CREATE'
        set_label_str = ""
        set_property_str = ""
        infer_corr_str = ""
        infer_observed_str = ""

        if node_constructor.set_labels is not None:
            set_label_str = f'SET $set_labels'''

        if len(node_constructor.result.optional_properties) > 0:
            set_property_str = 'SET $set_result_properties'

        if len(node_constructor.inferred_relationships) > 0:
            infer_corr_str = '''WITH record, $result_node_name'''
            for relationship in node_constructor.inferred_relationships:
                infer_rel_str = '''
                    CALL {WITH record, $result_node_name
                            MATCH ($event_node) - [:PREVALENCE] -> (record:$record_labels) <- [:PREVALENCE] - (
                                $result_node_name)
                                MERGE (event) - [:$relation_type] -> ($result_node_name)}'''
                infer_rel_str = Template(infer_rel_str).safe_substitute({
                    "event_node": relationship.event.get_pattern(name="event"),
                    "record_labels": relationship.get_labels_str(),
                    "relation_type": relationship.relation_type
                })
                infer_corr_str += infer_rel_str
        elif node_constructor.infer_corr_from_event_record:
            # language=SQL
            infer_corr_str = '''
            WITH record, $result_node_name
                                MATCH (event:$event_label) - [:PREVALENCE] -> (record) <- [:PREVALENCE] - (
                                $result_node_name)
                                MERGE (event) - [:$corr_type] -> ($result_node_name)'''
        elif node_constructor.infer_corr_from_entity_record:  # TODO update such that only correct events are considered
            # language=SQL
            infer_corr_str = '''
                        WITH record, $result_node_name
                                MATCH (event:$event_label) - [:PREVALENCE] -> (record) <- [:PREVALENCE] - (
                                $result_node_name)
                                MERGE (event) - [:$corr_type] -> ($result_node_name)'''
        elif node_constructor.infer_observed:
            # language=SQL
            infer_observed_str = '''
            WITH record, $result_node_name
                                MATCH (event:$event_label) - [:PREVALENCE] -> (record) <- [:PREVALENCE] - (
                                $result_node_name)
                                CREATE (event) <- [:OBSERVED] - ($result_node_name)
                                '''

        # language=SQL
        query_str = '''
                    CALL apoc.periodic.commit(
                        'MATCH ($record) 
                        WHERE NOT record:RecordCreated
                            WITH $record_name limit $limit
                            $merge_or_create ($result_node)
                            SET $record_name:RecordCreated
                            $set_label_str
                            $set_property_str
                            MERGE (record) <- [:PREVALENCE] - ($result_node_name)
                            $infer_corr_str
                            $infer_observed_str
                            RETURN count(*)',
                            {limit: $limit})
                    '''

        query_str = Template(query_str).safe_substitute({
            "set_label_str": set_label_str,
            "set_property_str": set_property_str,
            "infer_corr_str": infer_corr_str,
            "infer_observed_str": infer_observed_str,
            "merge_or_create": merge_or_create,
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "record": node_constructor.get_prevalent_record_pattern(node_name="record"),
                         "record_name": "record",
                         "conditions": node_constructor.get_where_condition(node_name="record", include_start_and=True),
                         "result_node": node_constructor.result.get_pattern(),
                         "result_node_name": node_constructor.result.get_name(),
                         "set_result_properties": node_constructor.get_set_result_properties_query(),
                         "set_labels": node_constructor.get_set_result_labels_query(),
                         "corr_type": node_constructor.corr_type,
                         "event_label": node_constructor.event_label
                     },
                     parameters={"limit": batch_size})

    @staticmethod
    def get_reset_created_record_query(batch_size: int):
        query_str = '''
                    CALL apoc.periodic.commit(
                        'MATCH (record:RecordCreated) 
                            WITH record limit $limit
                            REMOVE record:RecordCreated
                            RETURN count(*)',
                            {limit: $limit})
                    '''
        return Query(query_str=query_str,
                     parameters={"limit": batch_size})

    @staticmethod
    def get_number_of_ids_query(node_constructor: NodeConstructor, use_record: bool = False):
        query_str = '''MATCH (n:$labels)
                        RETURN count(DISTINCT n.sysId) as num_ids'''

        if use_record:
            query_str = '''MATCH ($record) 
                           RETURN count(DISTINCT record.$attribute) as num_ids'''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "labels": node_constructor.get_label_string(),
                         "attribute": node_constructor.result.required_properties[0].ref_attribute,
                         "record": node_constructor.get_prevalent_record_pattern(node_name="record")
                     })

    @staticmethod
    def get_number_of_records_query(node_constructor: NodeConstructor):

        query_str = '''MATCH ($record) 
                           RETURN count(record) as count'''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "record": node_constructor.get_prevalent_record_pattern(node_name="record")
                     })

    @staticmethod
    def get_reset_merged_in_nodes_query(node_constructor: NodeConstructor, batch_size: int):
        query_str = '''
                           CALL apoc.periodic.commit(
                                  'MATCH (n:$labels)
                                  WHERE n.merged = True
                                  WITH n LIMIT $limit
                                  SET n.merged = Null
                           RETURN COUNT(*)',
                           {limit:$limit})
                       '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "labels": node_constructor.get_label_string(),
                     },
                     parameters={"limit": batch_size})

    @staticmethod
    def get_merge_nodes_with_same_id_query(node_constructor: NodeConstructor, batch_size: int):
        if "Event" in node_constructor.get_labels():
            return None


        query_str = '''
                    CALL apoc.periodic.commit(
                           'MATCH (n:$labels)
                           WHERE n.merged IS NULL
                           // we order by sysId
                           WITH n ORDER BY n.sysId LIMIT $limit
                           WITH collect(n) as collection
                           WITH collection, last(collection) as last_node, size(collection) as collection_size
                           UNWIND collection as n
                           WITH $idt_properties, collect(n) as same_nodes, last_node, collection_size
                           CALL {WITH same_nodes, last_node
                                MATCH (last_node)
                                // last node could be the first of the list of same_nodes, we do not set to merged = true
                                // for this node
                                 WHERE size(same_nodes) = 1 and  head(same_nodes) <> last_node
                                 SET head(same_nodes).merged = True}             
                           WITH same_nodes, collection_size                   
                           WHERE size(same_nodes) > 1
                           UNWIND range(0, toInteger(floor(size(same_nodes)/100))) as i
                           WITH same_nodes, i*100 as min_range, apoc.coll.min([(i+1)*100, size(same_nodes)]) AS 
                           max_range, collection_size
                           WITH same_nodes[min_range..max_range] as lim_nodes, collection_size
                           CALL apoc.refactor.mergeNodes(lim_nodes, {properties: "discard", mergeRels: true})
                           YIELD node
                           RETURN collection_size',
                           {limit:$limit})
                       '''


        return Query(query_str=query_str,
                     template_string_parameters={
                         "labels": node_constructor.get_label_string(),
                         "idt_properties": node_constructor.get_idt_properties_query()
                     },
                     parameters={"limit": batch_size})



    @staticmethod
    def get_infer_corr_from_parent_query(batch_size, relation_constructor, use_from):
        if use_from:
            node =  relation_constructor.from_node.get_pattern()
            from_or_to = "FROM"
        else:
            node  = relation_constructor.to_node.get_pattern()
            from_or_to = "TO"

        query_str = '''
            CALL apoc.periodic.commit('
                MATCH (e:Event) --> ($node) - [:$from_or_to] - (relation:$relation_label_str)
                WHERE NOT EXISTS ((e) - [:CORR] -> (relation))
                WITH DISTINCT relation, e limit $limit
                MERGE (e) - [:CORR] -> (relation)
                RETURN COUNT(*)',
                {limit:$limit}
                )       
            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "node": node,
                         "from_or_to": from_or_to,
                         "relation_label_str": relation_constructor.result.get_relation_types_str()
                     },
                     parameters={
                         "limit": batch_size
                     })



    @staticmethod
    def get_create_relation_by_relations_query(relation_constructor: RelationConstructor, batch_size: int) -> Query:
        if relation_constructor.model_as_node:
            # language=sql
            merge_str = '''
                            MERGE ($from_node_name) -[:FROM] -> (relation:$relation_label_str) - [:TO] -> (
                            $to_node_name)
                            '''
        else:
            merge_str = "MERGE ($from_node_name) -[$rel_pattern] -> ($to_node_name)"

        # language=SQL
        query_str = '''
                CALL apoc.periodic.iterate(
                '$relation_queries                        
                RETURN distinct $from_node_name, $to_node_name',
                '$merge_str',                        
                {batchSize: $batch_size})
            '''

        query_str = Template(query_str).safe_substitute({
            "merge_str": merge_str
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "relation_queries": relation_constructor.get_relations_query(),
                         "from_node_name": relation_constructor.from_node.get_name(),
                         "to_node_name": relation_constructor.to_node.get_name(),
                         "rel_pattern": relation_constructor.result.get_pattern(),
                         "relation_label_str": relation_constructor.result.get_relation_types_str()
                     },
                     parameters={
                         "batch_size": batch_size
                     })

    @staticmethod
    def get_create_relation_using_record_query(relation_constructor: RelationConstructor, batch_size: int) -> Query:
        # find events that are related to different entities of which one event also has a reference to the other entity
        # create a relation between these two entities
        if relation_constructor.model_as_node:
            # language=sql
            merge_str = '''
                            MERGE ($from_node_name) -[:FROM] -> (relation:$relation_label_str) - [:TO] -> (
                            $to_node_name)
                            MERGE (relation)  - [:PREVALENCE] -> (record)
                            '''
        else:
            merge_str = "MERGE ($from_node_name) -[$rel_pattern] -> ($to_node_name)"

        query_str = '''     CALL apoc.periodic.commit('
                            MATCH (record:$record_labels)
                            WHERE NOT record:RecordCreated
                            WITH  record limit $limit
                            MATCH ($from_node) - [:PREVALENCE] -> (record)
                            MATCH ($to_node) - [:PREVALENCE] -> (record)
                            $merge_str
                            SET record:RecordCreated
                            RETURN COUNT(*)',
                            {limit:$limit})
                        '''

        query_str = Template(query_str).safe_substitute({
            "merge_str": merge_str
        })

        return Query(query_str=query_str,
                     template_string_parameters={
                         "from_node": relation_constructor.from_node.get_pattern(),
                         "from_node_name": relation_constructor.from_node.get_name(),
                         "to_node": relation_constructor.to_node.get_pattern(),
                         "to_node_name": relation_constructor.to_node.get_name(),
                         "record_labels": relation_constructor.prevalent_record.get_label_str(
                             include_first_colon=False),
                         "rel_pattern": relation_constructor.result.get_pattern("relation"),
                         "relation_labels": relation_constructor.result.get_relation_types_str(as_list=True),
                         "relation_label_str": relation_constructor.result.get_relation_types_str()
                     },
                     parameters={
                         "limit": batch_size
                     })

    @staticmethod
    def get_reset_created_relation_query(relation_constructor: RelationConstructor, batch_size: int) -> Query:
        # find events that are related to different entities of which one event also has a reference to the other entity
        # create a relation between these two entities

        query_str = '''     CALL apoc.periodic.commit('
                              MATCH (record:$record_labels)
                              WHERE record.rel_created = True
                              WITH record limit $limit
                              SET record.rel_created = Null
                              RETURN COUNT(*)',
                              {limit:$limit})
                          '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "record_labels": relation_constructor.prevalent_record.get_label_str(
                             include_first_colon=False),
                         "limit": batch_size
                     })

    @staticmethod
    def get_create_directly_follows_query(entity: ConstructedNodes, batch_size, event_label) -> Query:
        # find the specific entities and events with a certain label correlated to that entity
        # order all events by time, order_nr and id grouped by a node n
        # collect the sorted nodes as a list
        # unwind the list from 0 to the one-to-last node
        # find neighbouring nodes and add an edge between

        # language=sql

        if event_label == "CompoundEvent":
            if entity.node_type == "Resource":
                query_str = '''
                     CALL apoc.periodic.iterate(
                        'MATCH (n:$entity_labels_string) <-[:$corr_type_string]- (e:$event_label)
                        CALL {
                                WITH e
                                MATCH (e) - [:CONSISTS_OF] -> (single_event:Event)
                                RETURN id(single_event) as min_id ORDER BY id(single_event)
                                LIMIT 1
                            }
                        WITH n , e as nodes ORDER BY e.timestamp, min_id
                        WITH n , collect (nodes) as nodeList
                        UNWIND range(0,size(nodeList)-2) AS i
                        WITH n , nodeList[i] as first, nodeList[i+1] as second
                        RETURN n, first, second',
                        'MERGE (first) -[df:$df_entity {entityType: "$entity_type"}]->(second)
                         SET df.type = "DF"
                         SET df.entityId = n.sysId
                        ',
                        {batchSize: $batch_size})
                    '''
            else:
                query_str = '''
                                     CALL apoc.periodic.iterate(
                                        'MATCH (n:$entity_labels_string) <-[:$corr_type_string]- (e:$event_label)
                                        CALL {
                                                WITH e
                                                MATCH (e) - [:CONSISTS_OF] -> (single_event:Event)
                                                RETURN id(single_event) as min_id ORDER BY id(single_event)
                                                LIMIT 1
                                            }
                                        WITH n , e as nodes ORDER BY e.timestamp, min_id
                                        WITH n , collect (nodes) as nodeList
                                        UNWIND range(0,size(nodeList)-2) AS i
                                        WITH n , nodeList[i] as first, nodeList[i+1] as second
                                        RETURN first, second',
                                        'MERGE (first) -[df:$df_entity {entityType: "$entity_type"}]->(second)
                                         SET df.type = "DF"
                                        ',
                                        {batchSize: $batch_size})
                                    '''

        else:
            query_str = '''
                             CALL apoc.periodic.iterate(
                                'MATCH (n:$entity_labels_string) <-[:$corr_type_string]- (e:$event_label)
                                WITH n , e as nodes ORDER BY e.timestamp, ID(e)
                                WITH n , collect (nodes) as nodeList
                                UNWIND range(0,size(nodeList)-2) AS i
                                WITH n , nodeList[i] as first, nodeList[i+1] as second
                                RETURN first, second',
                                'MERGE (first) -[df:$df_entity {entityType: "$entity_type"}]->(second)
                                 SET df.type = "DF"
                                ',
                                {batchSize: $batch_size})
                            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_labels_string": entity.get_label_string(),
                         "corr_type_string": entity.get_corr_type_strings(),
                         "event_label": event_label,
                         "df_entity": entity.get_df_label(),
                         "entity_type": entity.node_type,
                         "batch_size": batch_size
                     })

    @staticmethod
    def get_merge_duplicate_df_entity_query(node: ConstructedNodes) -> Query:

        # language=sql
        query_str = '''
                        MATCH (n1:Event)-[r:$df_entity {entityType: '$entity_type'}]->(n2:Event)
                        WITH n1, n2, collect(r) AS rels
                        WHERE size(rels) > 1
                        // only include this and the next line if you want to remove the existing relationships
                        UNWIND rels AS r 
                        DELETE r
                        MERGE (n1)
                            -[:$df_entity {entityType: '$entity_type', count:size(rels), type: 'DF'}]->
                              (n2)
                    '''
        return Query(query_str=query_str,
                     template_string_parameters={
                         "entity_type": node.node_type,
                         "df_entity": node.get_df_label()
                     })

    @staticmethod
    def delete_parallel_directly_follows_derived(_type: str, node: Node):
        # language=sql
        query_str = '''
            MATCH (e1:Event) -[df:DF {entityType: "$type"}]-> (e2:Event)
            WHERE (e1:Event) -[:DF {entityType: "$original_entity_type"}]-> (e2:Event)
            DELETE df'''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "type": _type,
                         "original_entity_type": node.get_label_str()
                     },
                     parameters={})



from typing import Optional, List
from ..data_managers.semantic_header import ConstructedNodes
from ..database_managers.db_connection import Query


class AnalysisQueryLibrary:
    @staticmethod
    def get_dfc_label(entity_type: str, include_label_in_dfc: bool) -> str:
        if include_label_in_dfc:
            return f'DF_C_{entity_type.upper()}'
        else:
            return f'DF_C'

    @staticmethod
    def get_aggregate_df_relations_query(entity: ConstructedNodes,
                                         include_label_in_c_df: bool = True,
                                         df_threshold: int = 0,
                                         relative_df_threshold: float = 0,
                                         exclude_self_loops=True) -> Query:

        # add relations between classes when desired
        if df_threshold == 0 and relative_df_threshold == 0:
            # corresponds to aggregate_df_relations &  aggregate_df_relations_for_entities in graphdb-event-logs
            # aggregate only for a specific entity type and event classifier

            # language=sql
            query_str = '''
                                MATCH 
                                (c1:Activity) -[:OBSERVED]-> (e1:Event) 
                                    -[df:$df_label {entityType: '$entity_type'}]-> 
                                (e2:Event) <-  [:OBSERVED] - (c2:Activity)
                                $classifier_self_loops
                                WITH c1, count(df) AS df_freq,c2
                                MERGE 
                                (c1) 
                                    -[rel2:$dfc_label {entityType: '$entity_type', type:'DF_C'}]-> 
                                (c2) 
                                ON CREATE SET rel2.count=df_freq'''
        else:
            # aggregate only for a specific entity type and event classifier
            # include only edges with a minimum threshold, drop weak edges (similar to heuristics miner)

            # language=sql
            query_str = '''
                                MATCH 
                                (c1:Activity) 
                                    -[:OBSERVED]->
                                (e1:Event) 
                                    -[df:$df_label {entityType: '$entity_type'}]-> 
                                (e2:Event) <-[:OBSERVED]- (c2:Activity)
                                MATCH (e1) -[:CORR] -> (n) <-[:CORR]- (e2)
                                $classifier_self_loops
                                WITH c1,count(df) AS df_freq,c2
                                WHERE df_freq > $df_threshold
                                OPTIONAL MATCH (c2:Activity) -[:OBSERVED]-> (e2b:Event) -[df2:$df_label {entityType: '$entity_type'}]-> 
                                    (e1b:Event) <-[:OBSERVED]- (c2:Activity)
                                WITH c1,df_freq,count(df2) AS df_freq2,c2
                                WHERE (df_freq*$relative_df_threshold > df_freq2)
                                MERGE 
                                (c1) 
                                    -[rel2:$dfc_label {entityType: '$entity_type', type:'DF_C'}]-> 
                                (c2) 
                                ON CREATE SET rel2.count=df_freq'''


        return Query(query_str=query_str,
                     template_string_parameters={
                         "df_label": entity.get_df_label(),
                         "entity_type": entity.node_type,
                         "classifier_self_loops": "WHERE c1 <> c2" if exclude_self_loops else "",
                         "dfc_label": AnalysisQueryLibrary.get_dfc_label(entity.node_type, include_label_in_c_df),
                         "df_threshold": df_threshold,
                         "relative_df_threshold": relative_df_threshold
                     })



from ..cypher_queries.inference_engine_ql import InferenceEngineQueryLibrary as ie_ql
from ..utilities.performance_handling import Performance
from ..database_managers.db_connection import DatabaseConnection


class InferenceEngine:
    def __init__(self, db_connection: DatabaseConnection):
        self.connection = db_connection

    @Performance.track("entity")
    def infer_items_propagate_upwards_multiple_levels(self, entity, is_load=True):
        self.connection.exec_query(ie_ql.get_query_infer_items_propagate_upwards_multiple_levels,
                                   **{
                                       "entity": entity,
                                       "is_load": is_load
                                   })

    @Performance.track("entity")
    def match_entity_with_batch_position(self, entity, relative_position):
        self.connection.exec_query(ie_ql.get_match_entity_with_batch_position_query, **{
            "entity": entity,
            "relative_position": relative_position
        })

    @Performance.track("entity")
    def infer_items_propagate_downwards_multiple_level_w_batching(self, entity, relative_position):
        self.connection.exec_query(ie_ql.get_query_infer_items_propagate_downwards_multiple_level_w_batching,
                                   **{"entity": entity, "relative_position": relative_position})

    @Performance.track("entity")
    def infer_items_propagate_downwards_one_level(self, entity):
        self.connection.exec_query(ie_ql.get_query_infer_items_propagate_downwards_one_level, **{"entity": entity})



from typing import Dict, Optional, List, Union
from string import Template

from ..data_managers.datastructures import DataStructure
from ..data_managers.semantic_header import RecordConstructor
from ..database_managers.db_connection import Query


class DataImporterQueryLibrary:
    @staticmethod
    def get_optional_label_str(get_labels_str, optional_labels):
        record_name = "record"
        optional_label_str = ""
        if len(optional_labels) > 0:
            # language=SQL
            optional_label_str = '''
                    $assign_labels_str
                    CALL apoc.create.addLabels($record_name, $optional_labels_list)
                    YIELD node as record_with_labels
                    '''
        optional_label_str = Template(optional_label_str).safe_substitute({
            "record_name": record_name,
            "assign_labels_str": get_labels_str,
            "optional_labels_list": "+".join(optional_labels)
        })

        return optional_label_str

    @staticmethod
    def get_label_constructors(record_constructors: List[Dict[bool, RecordConstructor]]):
        required_labels = ['Record']
        optional_labels = []
        record_name = "record"
        get_labels_str = ""
        for index, record_constructor_dict in enumerate(record_constructors):
            record_constructor = record_constructor_dict["record_constructor"]
            if record_constructor_dict["required"]:
                required_labels.extend(record_constructor.record_labels)
            else:
                with_str = '''
                            CALL apoc.when($required_attributes_not_null $attributes_match, 'RETURN $labels as 
                            labels', 'RETURN [] as labels')
                            YIELD value as labels$i'''
                with_str = Template(with_str).safe_substitute({
                    "attributes_match": record_constructor.get_additional_conditions(),
                    "required_attributes_not_null": record_constructor.get_required_attributes_is_not_null_pattern(
                        record_name=record_name),
                    "labels": record_constructor.get_label_list(as_str=True),
                    "record_name": record_name,
                    "i": index
                })
                get_labels_str += with_str
                optional_labels.append(f"labels{index}['labels']")

        optional_labels_str = DataImporterQueryLibrary.get_optional_label_str(optional_labels=optional_labels,
                                                                              get_labels_str=get_labels_str)

        label_constructor = {
            "required_labels": required_labels,
            "optional_labels_str": optional_labels_str
        }

        return label_constructor

    @staticmethod
    def get_create_nodes_by_importing_batch_query(batch: List[Dict[str, str]],
                                                  labels_constructors: Dict[str, Union[List[str], str]]) -> Query:
        """
        Create event nodes for each row in the batch with labels
        The properties of each row are also the property of the node
        @param batch: List[Dictionary[key: value]], the key and its values form properties of the event nodes
        @param labels: The labels of the event nodes
        @return: None,
        """

        record_name = "record"
        required_labels = labels_constructors["required_labels"]
        optional_labels_str = labels_constructors["optional_labels_str"]

        # $batch is a variable we can add in tx.run, this allows us to use string properties
        # (keys in our dictionary are string)
        # return is required when using call and yield
        # language=SQL
        query_str = '''
                    UNWIND $batch AS row
                    CALL apoc.create.node($required_labels, row) YIELD node AS $record_name
                    $optional_label_str
                    RETURN count(*)
                '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "record_name": record_name,
                         "required_labels": required_labels,
                         "optional_label_str": optional_labels_str
                     },
                     parameters={"batch": batch})

    @staticmethod
    def get_make_timestamp_date_query(attribute, datetime_object, batch_size, load_status) -> Query:
        """
        Convert the strings of the timestamp to the datetime as used in cypher
        Remove the str_timestamp property
        @return: None
        """
        offset = datetime_object.timezone_offset
        offset = f'{attribute}+"{offset}"' if offset != "" else attribute

        # language=SQL
        query_str = '''
                CALL apoc.periodic.iterate(
                'MATCH (e:Record) WHERE e.$attribute IS NOT NULL AND e.loadStatus = $load_status 
                AND NOT apoc.meta.cypher.isType(e.$attribute, "$date_type")
                WITH e, e.$offset as timezone_dt
                WITH e, datetime(apoc.date.convertFormat(timezone_dt, "$datetime_object_format", 
                    "$datetime_object_convert_to")) as converted
                RETURN e, converted',
                'SET e.$attribute = converted',
                {batchSize:$batch_size, parallel:false})
            '''

        return Query(query_str=query_str,
                     template_string_parameters={
                         "datetime_object_format": datetime_object.format,
                         "datetime_object_convert_to": datetime_object.convert_to,
                         "date_type": datetime_object.get_date_type(),
                         "batch_size": batch_size,
                         "attribute": attribute,
                         "offset": offset,
                         "load_status": load_status
                     })

    @staticmethod
    def get_convert_epoch_to_timestamp_query(attribute, datetime_object, batch_size, load_status) -> Query:
        # language=SQL
        query_str = '''
                CALL apoc.periodic.iterate(
                'MATCH (e:Record) WHERE e.$attribute IS NOT NULL AND e.loadStatus = $load_status 
                WITH e, e.$attribute as timezone_dt
                WITH e, apoc.date.format(timezone_dt, $unit, 
                    $dt_format) as converted
                RETURN e, converted',
                'SET e.$attribute = converted',
                {batchSize:$batch_size, parallel:false, 
                params: {unit: $unit,
                        dt_format: $datetime_object_format}})
            '''

        return Query(query_str=query_str,
                     template_string_parameters={"attribute": attribute},
                     parameters={
                         "label": "Record",
                         "batch_size": batch_size,
                         "unit": datetime_object.unit,
                         "datetime_object_format": datetime_object.format,
                         "load_status": load_status
                     })

    @staticmethod
    def get_finalize_import_events_query(batch_size) -> Query:
        # language=SQL
        query_str = '''
            CALL apoc.periodic.iterate(
                'MATCH (e:Record) 
                WHERE e.loadStatus IS NOT NULL 
                RETURN e',
                'REMOVE e.loadStatus',
                {batchSize:$batch_size, parallel:false})
            '''

        return Query(query_str=query_str,
                     parameters={"batch_size": batch_size})

    @staticmethod
    def get_filter_events_by_property_query(prop: str, load_status: int, values: Optional[List[str]] = None,
                                            exclude=True) -> Query:
        if values is None:  # match all events that have a specific property
            negation = "NOT" if exclude else ""
            # query to delete all events and its relationship with property
            # language=SQL
            query_str = '''
                    MATCH (e:Event {loadStatus: $load_status})
                    WHERE e.$prop IS $negation NULL
                    DETACH DELETE e
                    '''
            template_string_parameters = {"prop": prop, "negation": negation}
        else:  # match all events with specific property and value
            negation = "" if exclude else "NOT"
            # match all e and delete them and its relationship
            # language=SQL
            query_str = '''
                    MATCH (e:Event {loadStatus: $load_status})
                    WHERE $negation e.$prop IN $values
                    DETACH DELETE e
                '''
            template_string_parameters = {
                "prop": prop, "negation": negation, "values": values, "load_status": load_status
            }

        # execute query
        return Query(query_str=query_str,
                     template_string_parameters=template_string_parameters)

    @staticmethod
    def get_update_load_status_query(current_load_status: int, batch_size: int):
        query_str = '''
        CALL apoc.periodic.commit(
            'MATCH (record:Record) 
            WHERE record.loadStatus = $old_value
            WITH record LIMIT $limit
            SET record.loadStatus = record.loadStatus + 1
            RETURN COUNT(*)',
            {old_value: $old_value,
            limit: $limit})
                            '''
        return Query(query_str=query_str,
                     parameters={
                         "old_value": current_load_status,
                         "limit": batch_size
                     })



from neo4j import GraphDatabase

driver = GraphDatabase.driver("bolt://localhost:7687", auth=("neo4j", "12345678"))

#match (e:Event),(n:Entity),(c:Class),(l:Log),()-[df:DF]->(),()-[corr:E_EN]->() return count(e),count(n),count(c),count(l),count(df),count(corr)

def get_size(tx):
    num_events, num_entity, num_log, num_class, num_df, num_corr, num_rel, num_total_nodes, num_total_relationships\
        = 0, 0, 0, 0, 0, 0, 0, 0, 0
    q = f'''match (e:Event) RETURN count(e)'''
    for record in tx.run(q):
        num_events = record["count(e)"]

    q = f'''match (e:Entity) RETURN count(e)'''
    for record in tx.run(q):
        num_entity = record["count(e)"]

    q = f'''match (l:Log) RETURN count(l)'''
    for record in tx.run(q):
        num_log = record["count(l)"]

    q = f'''match (c:Class) RETURN count(c)'''
    for record in tx.run(q):
        num_class = record["count(c)"]

    q = f'''match ()-[df:DF]->() RETURN count(df)'''
    for record in tx.run(q):
        num_df = record["count(df)"]

    q = f'''match ()-[corr:E_EN]->() RETURN count(corr)'''
    for record in tx.run(q):
        num_corr = record["count(corr)"]

    q = f'''match ()-[r:REL]->() RETURN count(r)'''
    for record in tx.run(q):
        num_rel = record["count(r)"]

    q = f'''match (n) RETURN count(n)'''
    for record in tx.run(q):
        num_total_nodes = record["count(n)"]

    q = f'''match ()-[r]->() RETURN count(r)'''
    for record in tx.run(q):
        num_total_relationships = record["count(r)"]

    print(f'''Nodes (total);Event;Entity;Log;Class;Relationships (total);DF;CORR;REL''')
    print(f'''{num_total_nodes};{num_events};{num_entity};{num_log};{num_class};{num_total_relationships};{num_df};
    {num_corr};{num_rel}''')



with driver.session() as session:
    session.read_transaction(get_size)



import re
from typing import Any, Optional, Dict, List


def replace_undefined_value(item, value):
    return item if item is not None else value


def create_list(class_type: Any, obj: Optional[Dict[str, Any]], *args) -> List[Any]:
    if obj is None:
        return []
    else:
        if type(obj) is list:
            new_list = [class_type.from_dict(y, *args) for y in obj]
        elif type(obj) is dict:
            new_list = [class_type.from_dict(obj, *args)]
        else:
            raise TypeError(f"The type of the object of {class_type} is ill defined")
        new_list = [item for item in new_list if item is not None]
        return new_list


def get_id_attribute_from_label(label):
    search_first_lower_case = re.search("[a-z]", label)
    if search_first_lower_case is not None:
        first_lower_case = search_first_lower_case.start()
    else: # no lower case has been found
        first_lower_case = len(label)
    return label[:first_lower_case].lower() + label[first_lower_case:] + "Id"



import json
import os
import re
import warnings
import random
from pathlib import Path

from typing import List, Dict, Any, Optional
from dataclasses import dataclass

import numpy as np
import pandas as pd
from pandas import DataFrame

from .semantic_header import Node
from ..utilities.auxiliary_functions import replace_undefined_value, create_list


@dataclass
class DatetimeObject:
    format: str
    timezone_offset: str
    convert_to: str
    is_epoch: bool
    unit: str

    def get_date_type(self):
        if self.convert_to == "ISO_DATE":
            return "DATE"
        else:
            return "DATE_TIME"

    @staticmethod
    def from_dict(obj: Any) -> 'DatetimeObject':
        if obj is None:
            return None
        _format = obj.get("format")
        _timezone_offset = replace_undefined_value(obj.get("timezone_offset"), "")
        _convert_to = "ISO_DATE"
        if re.search('[hkHK]', _format):
            _convert_to = "ISO_DATE_TIME"

        # _convert_to = str(obj.get("convert_to"))
        _is_epoch = replace_undefined_value(obj.get("is_epoch"), False)
        _unit = obj.get("unit")
        return DatetimeObject(_format, _timezone_offset, _convert_to, _is_epoch, _unit)


@dataclass
class Column:
    name: str
    dtype: str
    nan_values: List[str]
    optional: bool
    range_start: int
    range_end: int

    @staticmethod
    def from_dict(obj: Any) -> Optional['Column']:
        if obj is None:
            return None
        _name = obj.get("name")
        _dtype = obj.get("dtype")
        _nan_values = replace_undefined_value(obj.get("nan_values"), [])
        _optional = replace_undefined_value(obj.get("optional"), False)
        _range_start = obj.get("range_start")
        _range_end = obj.get("range_end")
        return Column(_name, _dtype, _nan_values, _optional, _range_start, _range_end)


@dataclass
class Attribute:
    name: str
    columns: List[Column]
    separator: str
    is_datetime: bool
    is_compound: bool
    optional: bool
    datetime_object: DatetimeObject
    na_rep_value: Any
    na_rep_columns: List[Column]
    filter_exclude_values: List[str]
    filter_include_values: List[str]
    use_filter: bool
    is_primary_key: bool
    is_foreign_key: bool

    @staticmethod
    def from_dict(obj: Any) -> Optional['Attribute']:
        if obj is None:
            return None
        _name = obj.get("name")
        _columns = create_list(Column, obj.get("columns"))
        _is_compound = len(_columns) > 1
        _optional = bool(obj.get("optional"))
        _datetime_object = DatetimeObject.from_dict(obj.get("datetime_object"))
        _is_datetime = _datetime_object is not None
        _na_rep_value = obj.get("na_rep_value")
        _na_rep_columns = create_list(Column, obj.get("na_rep_columns"))
        _separator = obj.get("separator")
        _filter_exclude_values = obj.get("filter_exclude_values")
        _filter_include_values = obj.get("filter_include_values")
        _use_filter = _filter_include_values is not None or _filter_exclude_values is not None  # default value
        _use_filter = replace_undefined_value(obj.get("use_filter"), _use_filter)
        _is_primary_key = replace_undefined_value(obj.get("is_primary_key"), False)
        _is_foreign_key = replace_undefined_value(obj.get("is_foreign_key"), False)
        return Attribute(name=_name, optional=_optional, columns=_columns, separator=_separator,
                         is_compound=_is_compound,
                         is_datetime=_is_datetime, datetime_object=_datetime_object,
                         na_rep_value=_na_rep_value, na_rep_columns=_na_rep_columns,
                         filter_exclude_values=_filter_exclude_values, filter_include_values=_filter_include_values,
                         use_filter=_use_filter, is_primary_key=_is_primary_key, is_foreign_key=_is_foreign_key)


@dataclass
class Sample:
    file_name: str
    use_random_sample: bool
    population_column: str
    size: int
    ids: List[Any]
    between: List[str]

    @staticmethod
    def from_dict(obj: Any, default_file_name: Optional[str] = None) -> Optional['Sample']:
        if obj is None:
            return None
        # when there is a single file specified, then we can use the default file name
        if default_file_name is not None:
            _file_name = replace_undefined_value(obj.get("file_name"), default_file_name)
        else:
            _file_name = obj.get("file_name")
        _use_random_sample = obj.get("use_random_sample")
        _population_column = obj.get("population_column")
        _size = obj.get("size")
        _ids = obj.get("ids")
        _between = obj.get("between")

        return Sample(_file_name, _use_random_sample, _population_column, _size, _ids, _between)


class DataStructure:
    def __init__(self, include: bool, name: str, file_directory: str, file_names: List[str],
                 encoding: str, seperator: str, decimal: str,
                 labels: List[str], true_values: List[str], false_values: List[str],
                 add_log: bool, add_index: bool,
                 samples: Dict[str, Sample], attributes: Dict[str, Attribute],
                 split_combined_events: bool):
        self.include = include
        self.name = name
        self.file_directory = file_directory
        self.file_names = file_names
        self.encoding = encoding
        self.seperator = seperator
        self.decimal = decimal
        self.labels = labels
        if self.labels == ["Record"]:
            self.labels = None
        self.true_values = true_values
        self.false_values = false_values
        self.add_log = add_log
        self.add_index = add_index
        self.samples = samples
        self.attributes = attributes
        self.split_combined_events = split_combined_events

    def __repr__(self):
        return self.name

    def has_datetime_attribute(self):
        return any([attribute.is_datetime for attribute in self.attributes.values()])
        # return "Event" in self.labels or "EventRecord" in self.labels

    def contains_composed_events(self):
        contains_composed_events = "startTimestamp" in self.get_datetime_formats() \
                                   or "completeTimestamp" in self.get_datetime_formats()
        return self.has_datetime_attribute() and contains_composed_events

    @staticmethod
    def from_dict(obj: Any) -> Optional['DataStructure']:
        if obj is None:
            return None

        _include = replace_undefined_value(obj.get("include"), True)

        if not _include:
            return None

        _name = obj.get("name")
        _path_of_executed_file = os.getcwd()
        _file_directory = os.path.join(os.getcwd(), *obj.get("file_directory").split("\\"))
        _file_names = obj.get("file_names") if obj.get("file_names") is not None else [obj.get("file_name")]
        _encoding = replace_undefined_value(obj.get("encoding"), "utf-8")
        _seperator = replace_undefined_value(obj.get("seperator"), ",")
        _decimal = replace_undefined_value(obj.get("decimal"), ".")
        _labels = obj.get("labels")
        _true_values = obj.get("true_values")
        _false_values = obj.get("false_values")
        _add_log = replace_undefined_value(obj.get("add_log"), False)
        _add_index = replace_undefined_value(obj.get("add_index"), True)

        _samples_obj = obj.get("samples") if obj.get("samples") is not None else obj.get("sample")
        if len(_file_names) == 1:  # single file name is defined
            _samples = create_list(Sample, _samples_obj, _file_names[0])
        else:
            _samples = create_list(Sample, _samples_obj)

        _samples = {sample.file_name: sample for sample in _samples}
        _attributes = create_list(Attribute, obj.get("attributes"))
        _attributes = {attribute.name: attribute for attribute in _attributes}
        _split_combined_events = replace_undefined_value(obj.get("split_combined_events"), False)

        return DataStructure(_include, _name, _file_directory, _file_names, _encoding, _seperator, _decimal,
                             _labels, _true_values, _false_values, _add_log, _add_index,
                             _samples, _attributes, _split_combined_events)

    def get_primary_keys(self):
        return [attribute_name for attribute_name, attribute in self.attributes.items() if attribute.is_primary_key]

    def get_primary_keys_as_attributes(self):
        primary_keys = self.get_primary_keys()
        primary_key_with = [f"n.{primary_key} as {primary_key}" for primary_key in primary_keys]
        primary_key_string = ", ".join(primary_key_with)
        return primary_key_string

    def get_label_string(self):
        return ":".join(self.labels)

    def get_foreign_keys(self):
        return [attribute_name for attribute_name, attribute in self.attributes.items() if attribute.is_foreign_key]

    def get_dtype_dict(self):
        dtypes = {}
        for attribute in self.attributes.values():
            for column in attribute.columns:
                if column.dtype is not None:
                    if column.name not in dtypes:
                        dtypes[column.name] = column.dtype
                    elif column.dtype != dtypes[column.name]:
                        warnings.warn(
                            f"Multiple dtypes ({column.dtype} != {dtypes[column.name]}) "
                            f"defined for {column.name}")
        return dtypes

    def get_required_columns(self):
        required_columns = set()
        for attribute in self.attributes.values():
            # add column names to the required columns
            required_columns.update([x.name for x in attribute.columns])
            required_columns.update([x.name for x in attribute.na_rep_columns])

        return list(required_columns)

    def create_sample(self, file_name, df_log):
        if self.samples is None or file_name not in self.samples:
            warnings.warn(f"No sample population has been defined for {self.name}")
            return df_log

        sample = self.samples[file_name]
        sample_column = sample.population_column
        if sample.use_random_sample:
            random_selection = random.sample(df_log[sample_column].unique().tolist(), k=sample.size)
            df_log = df_log[df_log[sample_column].isin(random_selection)]
        else:
            if sample.ids is not None:
                random_selection = sample.ids
                df_log = df_log[df_log[sample_column].isin(random_selection)]
            elif sample.between is not None:
                start_date = sample.between[0]
                end_date = sample.between[1]
                df_log['date_time_conv'] = pd.to_datetime(df_log[sample_column])
                df_log = df_log.loc[df_log["date_time_conv"].between(start_date, end_date)]
                df_log = df_log.drop(columns=['date_time_conv'])

        return df_log

    @staticmethod
    def replace_nan_values_based_on_na_rep_columns(df_log, attribute):
        if len(attribute.na_rep_columns) != len(attribute.columns):
            raise Exception(
                f"Na_rep_columns does not have the same size as columns for attribute {attribute.name}")
        else:  # they are the same size
            for i, na_rep_column in zip(range(len(attribute.na_rep_columns)), attribute.na_rep_columns):
                attribute_name = f"{attribute.name}_{i}"
                df_log[attribute_name].fillna(df_log[na_rep_column.name], inplace=True)

        return df_log

    @staticmethod
    def replace_nan_values_based_on_na_rep_value(df_log, attribute):
        for i in range(len(attribute.columns)):
            attribute_name = f"{attribute.name}_{i}"
            df_log[attribute_name].fillna(attribute.na_rep_value, inplace=True)

        return df_log

    @staticmethod
    def replace_nan_values_with_unknown(df_log, attribute):
        column: Column
        for i, column in zip(range(len(attribute.columns)), attribute.columns):
            attribute_name = f"{attribute.name}_{i}"
            if not column.optional:
                try:
                    df_log[attribute_name].fillna("Unknown", inplace=True)
                except:
                    df_log[attribute_name].fillna(-1, inplace=True)
        return df_log

    @staticmethod
    def create_compound_attribute(df_log, attribute):
        compound_column_names = [x.name for x in attribute.columns]
        df_log[attribute.name] = df_log[compound_column_names].apply(
            lambda row: attribute.separator.join([value for value in row.values.astype(str) if
                                                  not (value == 'nan' or value != value)]), axis=1)
        return df_log

    @staticmethod
    def combine_attribute_columns(df_log, attribute):
        compound_attribute_names = [f"{attribute.name}_{i}" for i in range(len(attribute.columns))]
        if attribute.is_compound:
            df_log[f"{attribute.name}_attribute"] = df_log[compound_attribute_names].apply(
                lambda row: attribute.separator.join([value for value in row.values.astype(str) if
                                                      not (value == 'nan' or value != value)]), axis=1)
        else:
            df_log[f"{attribute.name}_attribute"] = df_log[f"{attribute.name}_0"]
        df_log = df_log.drop(columns=compound_attribute_names)
        return df_log

    @staticmethod
    def create_attribute_columns(df_log, attribute):
        for i, column in zip(range(len(attribute.columns)), attribute.columns):
            attribute_name = f"{attribute.name}_{i}"
            df_log[attribute_name] = df_log[column.name]
            if column.range_start is not None or column.range_end is not None:
                df_log[attribute_name] = df_log[attribute_name].str[column.range_start:column.range_end]
        return df_log

    @staticmethod
    def replace_with_nan(df_log, attribute):
        for i, column in zip(range(len(attribute.columns)), attribute.columns):
            attribute_name = f"{attribute.name}_{i}"
            for nan_value in column.nan_values:
                df_log[attribute_name] = df_log[attribute_name].replace(nan_value, np.nan, regex=False)
        return df_log

    def preprocess_according_to_attributes(self, df_log):
        # loop over all attributes and check if they should be created, renamed or imputed
        for attribute in self.attributes.values():
            df_log = DataStructure.create_attribute_columns(df_log, attribute)
            df_log = DataStructure.replace_with_nan(df_log, attribute)
            if len(attribute.na_rep_columns) > 0:  # impute values in case of missing values
                df_log = DataStructure.replace_nan_values_based_on_na_rep_columns(df_log, attribute)
            if attribute.na_rep_value is not None:
                df_log = DataStructure.replace_nan_values_based_on_na_rep_value(df_log, attribute)

            df_log = DataStructure.combine_attribute_columns(df_log, attribute)

        return df_log

    def split_df_log_into_combined_events(self, df_log: DataFrame):
        df_log["idx"] = df_log.reset_index().index
        if "timestamp" in self.attributes:
            raise ImportError(
                "Combined events cannot be split since an attribute with name timestamp is already defined")

        if "startTimestamp" in self.attributes and "completeTimestamp" in self.attributes:
            df_log_start = df_log.drop(columns=["completeTimestamp"])
            df_log_start["lifecycle"] = "start"
            df_log_start = df_log_start.rename(columns={"startTimestamp": "timestamp"})

            df_log_end = df_log.drop(columns=["startTimestamp"])
            df_log_end["lifecycle"] = "complete"
            df_log_end = df_log_end.rename(columns={"completeTimestamp": "timestamp"})

            df_log = pd.concat([df_log_start, df_log_end])
        elif "startTimestamp" in self.attributes:
            df_log["lifecycle"] = "start"
            df_log = df_log.rename(columns={"startTimestamp": "timestamp"})
        elif "completeTimestamp" in self.attributes:
            df_log["lifecycle"] = "complete"
            df_log = df_log.rename(columns={"completeTimestamp": "timestamp"})

        df_log = df_log.sort_values(by=["timestamp", "idx"])
        df_log = df_log.drop(columns=["idx"])
        return df_log

    def update_attributes(self):
        if "startTimestamp" in self.attributes and "completeTimestamp" in self.attributes:
            start_dt_format = self.get_datetime_formats()["startTimestamp"].format
            complete_dt_format = self.get_datetime_formats()["completeTimestamp"].format
            if start_dt_format != complete_dt_format:
                raise ValueError("startTimestamp and completeTimestamp have a different format")

            start_attribute = self.attributes["startTimestamp"]
            start_attribute.name = "timestamp"
            self.attributes["timestamp"] = start_attribute
            del self.attributes["startTimestamp"]
            del self.attributes["completeTimestamp"]


        elif "startTimestamp" in self.attributes:
            start_attribute = self.attributes["startTimestamp"]
            start_attribute.name = "timestamp"

            self.attributes["timestamp"] = start_attribute
            del self.attributes["startTimestamp"]
        elif "completeTimestamp" in self.attributes:
            complete_attribute = self.attributes["completeTimestamp"]
            complete_attribute.name = "timestamp"

            self.attributes["timestamp"] = complete_attribute
            del self.attributes["completeTimestamp"]

    def prepare_event_data_sets(self, input_path, file_name, use_sample):
        dtypes = self.get_dtype_dict()
        required_columns = self.get_required_columns()

        true_values = self.true_values
        false_values = self.false_values

        if file_name.endswith('.csv'):
            df_log: DataFrame = pd.read_csv(os.path.join(input_path, file_name), keep_default_na=True,
                                            usecols=required_columns, dtype=dtypes, true_values=true_values,
                                            false_values=false_values, sep=self.seperator, decimal=self.decimal,
                                            encoding=self.encoding)
        else:
            raise TypeError(f"The file extension of {file_name} is not implemented. Use .csv.")

        if use_sample and self.has_datetime_attribute():
            df_log = self.create_sample(file_name, df_log)

        df_log = self.preprocess_according_to_attributes(df_log)

        # all columns have been renamed to or constructed with the name attribute,
        # hence only keep those that match with a name attribute
        required_attributes = list([f"{attribute_name}_attribute" for attribute_name in self.attributes.keys()])
        required_attributes_mapping = {f"{attribute_name}_attribute": f"{attribute_name}" for attribute_name in
                                       self.attributes.keys()}
        df_log = df_log[required_attributes]
        df_log = df_log.rename(columns=required_attributes_mapping)

        if self.split_combined_events:
            df_log = self.split_df_log_into_combined_events(df_log)
            self.update_attributes()

        if self.add_log:
            df_log["log"] = file_name

        if self.add_index:
            df_log["index"] = df_log.index

        # drop all columns with only nan values
        df_log = df_log.dropna(how='all', axis=1)  # drop all columns in which all values are nan (empty)
        df_log = df_log.dropna(how='all')  # drop all columns in which all values are nan (empty)

        return df_log

    def read_data_set(self, input_path, file_name, use_sample, store_preprocessed_file=True,
                      use_preprocessed_file=False):

        preprocessed_file_directory = os.path.join(input_path, "preprocessed_files")
        # change extension from csv to pkl and add sample in case of sample
        sample_is_used = use_sample and len(self.samples) > 0
        preprocessed_file_name = f"{file_name[:-4]}_sample.pkl" if sample_is_used else f"{file_name[:-4]}.pkl"
        preprocessed_file_path = os.path.join(preprocessed_file_directory, preprocessed_file_name)
        preprocessed_file_is_used = False
        if not use_preprocessed_file:
            df_log = self.prepare_event_data_sets(input_path, file_name, use_sample)
        elif not os.path.exists(preprocessed_file_path):
            warning_message = f"No preprocessed file {preprocessed_file_name} found, preprocessed the file instead"
            warnings.warn(warning_message)
            df_log = self.prepare_event_data_sets(input_path, file_name, use_sample)
        else:  # use_preprocessed_file and file exists
            df_log = pd.read_pickle(preprocessed_file_path)
            preprocessed_file_is_used = True

        if store_preprocessed_file and not preprocessed_file_is_used:
            os.makedirs(preprocessed_file_directory, exist_ok=True)
            df_log.to_pickle(preprocessed_file_path)

        return df_log

    def get_datetime_formats(self) -> Dict[str, DatetimeObject]:
        datetime_formats = {}

        for attribute_name, attribute in self.attributes.items():
            if attribute.is_datetime:
                datetime_formats[attribute_name] = attribute.datetime_object

        return datetime_formats

    def get_attribute_value_pairs_filtered(self, exclude: bool = True) -> Dict[str, List[str]]:
        attribute_value_pairs = {}

        for attribute_name, attribute in self.attributes.items():
            if attribute.use_filter:
                attribute_value_pairs[attribute_name] \
                    = attribute.filter_exclude_values if exclude else attribute.filter_include_values

        return attribute_value_pairs


class ImportedDataStructures:
    def __init__(self, path: Path):
        random.seed(1)
        with open(path, encoding='utf-8') as f:
            json_event_tables = json.load(f)

        self.structures = [DataStructure.from_dict(item) for item in json_event_tables]
        self.structures = [item for item in self.structures if item is not None]



from typing import Optional, List

from ..data_managers.semantic_header import ConstructedNodes, ConstructedRelation, Relationship, SemanticHeader, \
    NodeConstructor
from ..database_managers.db_connection import DatabaseConnection
from ..utilities.performance_handling import Performance
from ..cypher_queries.semantic_header_ql import SemanticHeaderQueryLibrary as sh_ql


class EKGUsingSemanticHeaderBuilder:
    def __init__(self, db_connection: DatabaseConnection, semantic_header: SemanticHeader, batch_size: int):
        self.connection = db_connection
        self.semantic_header = semantic_header
        self.batch_size = batch_size

    def create_nodes_by_records(self, node_types: Optional[List[str]]) -> None:
        for node_constructor in self.semantic_header.get_node_by_record_constructors(node_types):
            self._create_node_by_record(node_constructor=node_constructor)

    @Performance.track("node_constructor")
    def _create_node_by_record(self, node_constructor: NodeConstructor):
        num_ids = self.connection.exec_query(sh_ql.get_number_of_ids_query,
                                             **{
                                                 "node_constructor": node_constructor,
                                                 "use_record": True
                                             })
        merge_first = num_ids[0]['num_ids'] < 1000 \
                        and "Event" not in node_constructor.get_labels() \
                        and "EntityAttribute" not in node_constructor.get_labels()

        self.connection.exec_query(sh_ql.get_create_node_by_record_constructor_query,
                                   **{
                                       "node_constructor": node_constructor,
                                       "batch_size": self.batch_size,
                                       "merge": merge_first
                                   })

        self.connection.exec_query(sh_ql.get_reset_created_record_query,
                                   **{
                                       "batch_size": self.batch_size
                                   }
                                   )

        if merge_first:
            print(
                f"Node ({node_constructor.get_pattern(with_properties=False)})"
                f"using ({node_constructor.get_prevalent_record_pattern()}) merged")
        else:
            print(f"Node ({node_constructor.get_pattern(with_properties=False)}) "
                                               f"using ({node_constructor.get_prevalent_record_pattern()}) "
                                               f"created")
            if not ("Event" in node_constructor.get_labels() or "EntityAttribute" in node_constructor.get_labels()):
                max_limit = self.connection.exec_query(sh_ql.get_number_of_ids_query,
                                                       **{"node_constructor": node_constructor})
                self.connection.exec_query(sh_ql.get_merge_nodes_with_same_id_query,
                                           **{
                                               "node_constructor": node_constructor,
                                               "batch_size": self.batch_size
                                           }
                                           )

                self.connection.exec_query(sh_ql.get_reset_merged_in_nodes_query,
                                           **{
                                               "node_constructor": node_constructor,
                                               "batch_size": self.batch_size}
                                           )




    def create_nodes_by_relations(self, node_types: Optional[List[str]]) -> None:
        for node_constructors in self.semantic_header.get_nodes_constructed_by_relations(node_types).values():
            for node_constructor in node_constructors:
                self._create_node_by_record(node_constructor=node_constructor)



    def create_relations_using_records(self, relation_types: Optional[List[str]]) -> None:
        # find events that are related to different entities of which one event also has a reference to the other entity
        # create a relation between these two entities
        relation: ConstructedRelation
        for relation_constructor in self.semantic_header.get_relations_constructed_by_record(relation_types):
            self._create_relations_using_record(relation_constructor=relation_constructor)

    @Performance.track("relation_constructor")
    def _create_relations_using_record(self, relation_constructor):
        self.connection.exec_query(sh_ql.get_create_relation_using_record_query,
                                   **{
                                       "relation_constructor": relation_constructor,
                                       "batch_size": self.batch_size
                                   })
        self.connection.exec_query(sh_ql.get_reset_created_record_query,
                                   **{
                                       "batch_size": self.batch_size
                                   })
        self._create_corr_from_parents(relation_constructor=relation_constructor)

    def create_relations_using_relations(self, relation_types: Optional[List[str]]) -> None:
        relation: ConstructedRelation
        for relation_constructor in self.semantic_header.get_relations_constructed_by_relations(relation_types):
            self._create_relations_using_relation(relation_constructor=relation_constructor)

    @Performance.track("relation_constructor")
    def _create_relations_using_relation(self, relation_constructor):
        self.connection.exec_query(sh_ql.get_create_relation_by_relations_query,
                                   **{
                                       "relation_constructor": relation_constructor,
                                       "batch_size": self.batch_size
                                   })
        self._create_corr_from_parents(relation_constructor=relation_constructor)


    def _create_corr_from_parents(self, relation_constructor):
        if relation_constructor.infer_corr_from_reified_parents:
            for use_from in [True, False]:
                self.connection.exec_query(sh_ql.get_infer_corr_from_parent_query,
                                           **{
                                               "relation_constructor": relation_constructor,
                                               "use_from": use_from,
                                               "batch_size": self.batch_size
                                           })


    def create_df_edges(self, entity_types: List[str], event_label: str) -> None:
        entity: ConstructedNodes

        if entity_types is None:
            entity_types = [entity.node_type for entity in self.semantic_header.nodes]

        for entity in self.semantic_header.nodes:
            if entity.infer_df and entity.node_type in entity_types:
                self._create_df_edges_for_entity(entity=entity, event_label=event_label)

    @Performance.track("entity")
    def _create_df_edges_for_entity(self, entity: ConstructedNodes, event_label):
        self.connection.exec_query(sh_ql.get_create_directly_follows_query,
                                   **{
                                       "entity": entity, "batch_size": self.batch_size,
                                       "event_label": event_label
                                   })

    def merge_duplicate_df(self):
        node: ConstructedNodes
        for node in self.semantic_header.nodes:
            if node.merge_duplicate_df:
                self._merge_duplicate_df_for_node(node=node)

    @Performance.track("node")
    def _merge_duplicate_df_for_node(self, node: ConstructedNodes):
        self.connection.exec_query(sh_ql.get_merge_duplicate_df_entity_query, **{"node": node})

    def delete_parallel_dfs_derived(self):
        node: ConstructedNodes
        original_entity: ConstructedNodes
        relation: Relationship
        node_constructor: NodeConstructor
        for _type, node_constructor in self.semantic_header.get_nodes_constructed_by_relations(
                only_include_delete_parallel_df=True).items():
            self._delete_parallel_dfs_derived_for_node(node=node_constructor, type=_type)

    @Performance.track("type")
    def _delete_parallel_dfs_derived_for_node(self, node: NodeConstructor, type: str):
        from_node = node.relation.from_node
        to_node = node.relation.to_node
        for node in [from_node, to_node]:
            self.connection.exec_query(sh_ql.delete_parallel_directly_follows_derived,
                                       **{
                                           "type": type,
                                           "node": node
                                       })

    def create_static_nodes_and_relations(self):
        print("No implementation yet")
        pass



import os
import sys
import time
from datetime import datetime

import pandas as pd
from tqdm import tqdm

from ..utilities.context_manager_tqdm import Nostdout


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Performance(metaclass=Singleton):
    def __init__(self, perf_path: str, number_of_steps: int = None):
        self.start = time.time()
        self.last = self.start
        self.perf = pd.DataFrame(columns=["name", "start", "end", "duration"])
        self.path = perf_path
        self.count = 0
        if number_of_steps is not None:
            self.pbar = tqdm(total=number_of_steps, file=sys.stdout)
        else:
            self.pbar = tqdm(file=sys.stdout)
        self.total = None
        # start python trickery
        self.ctx = Nostdout()
        self.ctx.__enter__()

    def string_time(self, epoch_time):
        return datetime.utcfromtimestamp(epoch_time).strftime("%H:%M:%S")

    def finished_step(self, log_message: str = None):
        end = time.time()
        if log_message is None:
            self.perf = pd.concat([self.perf, pd.DataFrame.from_records([
                {
                    "name": log_message,
                    "start": self.string_time(self.last),
                    "end": self.string_time(end),
                    "duration": (end - self.last)
                }])])
        self.pbar.set_description(f"{log_message}: took {round(end - self.last, 2)} seconds")
        self.last = end
        self.count += 1
        self.pbar.update(1)

    def track(argument: str = None):
        def performance_tracker_wrapper(func):
            def wrapper(self, *args, **kwargs):
                func(self, *args, **kwargs)
                end = time.time()
                perf = Performance()
                if argument is None or argument not in kwargs:
                    log_message = func.__name__
                else:
                    log_message = f"{func.__name__} for {kwargs[argument]}"
                perf.perf = pd.concat([perf.perf, pd.DataFrame.from_records([
                    {
                        "name": log_message,
                        "start": perf.string_time(perf.last),
                        "end": perf.string_time(end),
                        "duration": (end - perf.last)
                    }])])
                perf.pbar.set_description(f"{log_message}: took {round(end - perf.last, 2)} seconds")
                perf.last = end
                perf.count += 1
                perf.pbar.update(1)

            return wrapper

        return performance_tracker_wrapper

    def finish(self):
        end = time.time()
        print(f"{self.count} steps")
        self.perf = pd.concat([self.perf, pd.DataFrame.from_records([
            {
                "name": "total",
                "start": self.string_time(self.start),
                "end": self.string_time(end),
                "duration": (end - self.start)
            }])])
        self.total = round(end - self.start, 2)
        print(f"Total: took {round(end - self.start, 2)} seconds")
        self.pbar.set_description(f"Completed")
        self.pbar.close()
        # close python trickery
        self.ctx.__exit__()

    def save(self):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        self.perf.to_csv(self.path, sep=";",decimal=",")
